"""DSW Template Development Kit

Template Development Kit for `Data Stewardship Wizard`_.

.. _Data Stewardship Wizard:
   https://ds-wizard.org

"""
from . import consts
from .cli import main


__app__ = consts.APP
__version__ = consts.VERSION

__all__ = ['__app__', '__version__', 'main']



import asyncio
import datetime
import io
import json
import logging
import pathlib
import shutil
import tempfile
import zipfile

import watchfiles

from . import consts
from .api_client import WizardAPIClient, WizardCommunicationError
from .model import Template, TemplateFile, TemplateFileType, TemplateProject
from .utils import UUIDGen
from .validation import TemplateValidator, ValidationError


ChangeItem = tuple[watchfiles.Change, pathlib.Path]


def _change(item: ChangeItem, root: pathlib.Path) -> str:
    return f'{item[0].name.upper()}[{item[1].relative_to(root).as_posix()}]'


class TDKProcessingError(RuntimeError):

    def __init__(self, message: str, hint: str):
        self.message = message
        self.hint = hint


# pylint: disable-next=too-many-public-methods
class TDKCore:

    # pylint: disable-next=too-many-locals
    def _check_metamodel_version(self):
        hint = 'Fix your metamodelVersion in template.json and/or visit docs'
        if self.remote_metamodel_version is None:
            self.logger.warning('Remote metamodel version is unknown, you should use '
                                'a matching version of TDK to avoid issues')
            return
        mm_ver = str(self.safe_template.metamodel_version)
        try:
            if '.' not in mm_ver:
                mm_ver = f'{mm_ver}.0'
            mm_major, mm_minor = map(int, mm_ver.split('.'))
        except ValueError as e:
            raise TDKProcessingError(f'Invalid metamodel version format: {mm_ver}', hint) from e
        mmr_ver = self.remote_metamodel_version
        try:
            if '.' not in mmr_ver:
                mmr_ver = f'{mmr_ver}.0'
            mmr_major, mmr_minor = map(int, mmr_ver.split('.'))
        except ValueError as e:
            raise TDKProcessingError(f'Invalid remote metamodel version format: {mmr_ver}',
                                     'Check if connecting to correct API with matching version '
                                     'as you have in TDK.') from e
        if (mm_major, mm_minor) == (mmr_major, mmr_minor):
            self.logger.debug('Metamodel version %s matches remote version %s',
                              mmr_ver, mmr_ver)
        elif mm_major == mmr_major and mm_minor < mmr_minor:
            self.logger.warning('Local metamodel version %s is older than remote version %s, '
                                'but still compatible', mm_ver, mmr_ver)
        else:
            raise TDKProcessingError(
                f'Unsupported metamodel version: local {mm_ver}, remote {mmr_ver}',
                hint,
            )

    def __init__(self, template: Template | None = None, project: TemplateProject | None = None,
                 client: WizardAPIClient | None = None, logger: logging.Logger | None = None):
        self.template = template
        self.project = project
        self.client = client
        self.remote_version: str = 'unknown~??????'
        self.remote_metamodel_version: str | None = 'unknown'
        self.logger = logger or logging.getLogger()
        self.loop = asyncio.get_event_loop()
        self.changes_processor = ChangesProcessor(self)
        self.remote_id = 'unknown'
        self.remote_uuid = None

    async def close(self):
        await self.safe_client.close()

    @property
    def remote_editor_url(self) -> str:
        if self.remote_uuid is None:
            raise RuntimeError('Remote template draft is not linked (yet)')
        return f'{self.safe_client.api_url}/document-template-editors/{self.remote_uuid}'

    @property
    def safe_template(self) -> Template:
        if self.template is None:
            raise RuntimeError('No template is loaded')
        return self.template

    @property
    def safe_project(self) -> TemplateProject:
        if self.project is None:
            raise RuntimeError('No template is loaded')
        return self.project

    @property
    def safe_client(self) -> WizardAPIClient:
        if self.client is None:
            raise RuntimeError('No DSW API client specified')
        return self.client

    async def init_client(self, api_url: str, api_key: str):
        self.logger.info('Connecting to %s', api_url)
        self.client = WizardAPIClient(api_url=api_url, api_key=api_key)
        self.remote_version, self.remote_metamodel_version = await self.client.get_api_version()
        user = await self.client.get_current_user()
        self.logger.info('Successfully authenticated as %s %s (%s)',
                         user['firstName'], user['lastName'], user['email'])
        self.logger.debug('Connected to API version %s', self.remote_version)

    def prepare_local(self, template_dir):
        self.logger.debug('Preparing local template project')
        self.project = TemplateProject(template_dir=template_dir, logger=self.logger)

    def load_local(self, template_dir):
        self.prepare_local(template_dir=template_dir)
        self.logger.info('Loading local template project')
        self.safe_project.load()

    async def _get_template_uuid(self, template_id: str) -> str:
        self.logger.debug('Finding template %s', template_id)
        t_uuid = await self.safe_client.find_template_uuid_by_id(template_id=template_id)
        if t_uuid is None:
            raise RuntimeError(f'Template with id {template_id} not found on remote')
        return t_uuid

    async def _get_template_draft_uuid(self, template_id: str) -> str:
        self.logger.debug('Finding template draft %s', template_id)
        t_uuid = await self.safe_client.find_template_draft_uuid_by_id(template_id=template_id)
        if t_uuid is None:
            raise RuntimeError(f'Template draft with id {template_id} not found on remote')
        return t_uuid

    async def load_remote(self, template_id: str):
        template_uuid = await self._get_template_draft_uuid(template_id=template_id)
        self.logger.info('Retrieving template draft %s (%s)',
                         template_uuid, template_id)
        self.template = await self.safe_client.get_template_draft(remote_uuid=template_uuid)
        if self.template.coordinates != template_id:
            self.logger.warning('Template coordinates %s do not match requested id %s',
                                self.template.coordinates, template_id)
        self.logger.debug('Retrieving template draft files')
        files = await self.safe_client.get_template_draft_files(remote_uuid=template_uuid)
        self.logger.info('Retrieved %s file(s)', len(files))
        for file in files:
            self.safe_template.files[file.filename.as_posix()] = file
        self.logger.debug('Retrieving template draft assets')
        assets = await self.safe_client.get_template_draft_assets(remote_uuid=template_uuid)
        self.logger.info('Retrieved %s asset(s)', len(assets))
        for asset in assets:
            self.safe_template.files[asset.filename.as_posix()] = asset

    async def download_bundle(self, template_id: str) -> bytes:
        template_uuid = await self._get_template_uuid(template_id=template_id)
        self.logger.info('Retrieving template %s (%s) bundle',
                         template_uuid, template_id)
        return await self.safe_client.get_template_bundle(remote_uuid=template_uuid)

    async def list_remote_templates(self) -> list[Template]:
        self.logger.info('Listing remote document templates')
        return await self.safe_client.get_templates()

    async def list_remote_drafts(self) -> list[Template]:
        self.logger.info('Listing remote document template drafts')
        return await self.safe_client.get_drafts()

    def verify(self) -> list[ValidationError]:
        template = self.template or self.safe_project.template
        if template is None:
            raise RuntimeError('No template is loaded')
        return TemplateValidator.collect_errors(template)

    def store_local(self, force: bool):
        if self.project is None:
            raise RuntimeError('No template project is initialized')
        self.project.template = self.safe_template
        if self.project.template is None:
            raise RuntimeError('No template is loaded in the project')
        if len(self.project.safe_template.tdk_config.files) == 0:
            self.project.safe_template.tdk_config.use_default_files()
            self.logger.warning('Using default _tdk.files in template.json, you may want '
                                'to change it to include relevant files')
        self.logger.debug('Initiating storing local template project (force=%s)', force)
        self.project.store(force=force)

    async def store_remote(self, force: bool):
        self.template = self.safe_project.template
        self._check_metamodel_version()
        org_id = await self.safe_client.get_organization_id()
        if org_id != self.safe_template.organization_id:
            self.logger.warning('There is different organization ID set in the DSW instance'
                                ' (local: %s, remote: %s)',
                                self.safe_template.organization_id, org_id)
        self.remote_id = self.safe_template.id_with_org(org_id)
        self.remote_uuid = await self.safe_client.find_template_draft_uuid_by_id(
            template_id=self.remote_id,
        )
        template_exists = self.remote_uuid is not None
        if template_exists and force:
            self.logger.warning('Deleting existing remote document template draft (forced)')
            result = await self.safe_client.delete_template_draft(remote_uuid=self.remote_uuid)
            if not result:
                self.logger.error('Could not delete document template draft')
            template_exists = not result

        if template_exists:
            self.logger.info('Updating existing remote document template draft')
            await self.safe_client.update_template_draft(
                template=self.safe_template,
                remote_uuid=self.remote_uuid,
            )
            self.logger.debug('Retrieving remote assets')
            remote_assets = await self.safe_client.get_template_draft_assets(
                remote_uuid=self.remote_uuid,
            )
            self.logger.debug('Retrieving remote files')
            remote_files = await self.safe_client.get_template_draft_files(
                remote_uuid=self.remote_uuid,
            )
            await self.cleanup_remote_files(
                remote_assets=remote_assets,
                remote_files=remote_files,
            )
            self.logger.info('Using existing editor with UUID: %s', self.remote_uuid)
        else:
            self.logger.info('Creating remote document template draft')
            result = await self.safe_client.create_new_template_draft(
                template=self.safe_template,
                remote_id=self.remote_id,
            )
            self.remote_uuid = result.uuid
            if result is None:
                raise RuntimeError('Failed to create document template draft')
            self.logger.info('Using new editor with UUID: %s', self.remote_uuid)
        self.logger.info('==> URL: %s', self.remote_editor_url)
        await self.store_remote_files()

    async def _update_template_file(self, remote_file: TemplateFile, local_file: TemplateFile,
                                    project_update: bool = False):
        try:
            self.logger.debug('Updating existing remote %s %s (%s) started',
                              remote_file.remote_type.value, remote_file.filename.as_posix(),
                              remote_file.remote_id)
            local_file.remote_id = remote_file.remote_id
            if remote_file.remote_type == TemplateFileType.ASSET:
                result = await self.safe_client.put_template_draft_asset_content(
                    remote_uuid=self.remote_uuid,
                    file=local_file,
                )
            else:
                result = await self.safe_client.put_template_draft_file_content(
                    remote_uuid=self.remote_uuid,
                    file=local_file,
                )
            self.logger.debug('Updating existing remote %s %s (%s) finished: %s',
                              remote_file.remote_type.value, remote_file.filename.as_posix(),
                              remote_file.remote_id, 'ok' if result else 'failed')
            if project_update and result:
                self.safe_project.update_template_file(result)
        except Exception as e1:
            try:
                self.logger.debug('Trying to delete/create due to: %s', str(e1))
                await self._delete_template_file(file=remote_file)
                await self._create_template_file(file=local_file, project_update=True)
            except Exception as e2:
                self.logger.error('Failed to update existing remote %s %s: %s',
                                  remote_file.remote_type.value,
                                  remote_file.filename.as_posix(), e2)

    async def _delete_template_file(self, file: TemplateFile, project_update: bool = False):
        try:
            self.logger.debug('Deleting existing remote %s %s (%s) started',
                              file.remote_type.value, file.filename.as_posix(),
                              file.remote_id)
            if file.remote_type == TemplateFileType.ASSET:
                result = await self.safe_client.delete_template_draft_asset(
                    remote_uuid=self.remote_uuid,
                    asset_uuid=file.remote_id,
                )
            else:
                result = await self.safe_client.delete_template_draft_file(
                    remote_uuid=self.remote_uuid,
                    file_uuid=file.remote_id,
                )
            self.logger.debug('Deleting existing remote %s %s (%s) finished: %s',
                              file.remote_type.value,
                              file.filename.as_posix(),
                              file.remote_id, 'ok' if result else 'failed')
            if project_update and result:
                self.safe_project.remove_template_file(file.filename)
        except Exception as e:
            self.logger.error('Failed to delete existing remote %s %s: %s',
                              file.remote_type.value, file.filename.as_posix(), e)

    async def cleanup_remote_files(self, remote_assets: list[TemplateFile],
                                   remote_files: list[TemplateFile]):
        for local_file in self.safe_project.safe_template.files.values():
            self.logger.debug('Cleaning up remote %s', local_file.filename.as_posix())
            for remote_asset in remote_assets:
                if remote_asset.filename == local_file.filename:
                    await self._delete_template_file(file=remote_asset, project_update=False)
            for remote_file in remote_files:
                if remote_file.filename == local_file.filename:
                    await self._delete_template_file(file=remote_file, project_update=False)

    async def _create_template_file(self, file: TemplateFile, project_update: bool = False):
        try:
            self.logger.debug('Storing remote %s %s started',
                              file.remote_type.value, file.filename.as_posix())
            if file.remote_type == TemplateFileType.ASSET:
                result = await self.safe_client.post_template_draft_asset(
                    remote_uuid=self.remote_uuid,
                    file=file,
                )
            else:
                result = await self.safe_client.post_template_draft_file(
                    remote_uuid=self.remote_uuid,
                    file=file,
                )
            self.logger.debug('Storing remote %s %s finished: %s',
                              file.remote_type.value, file.filename.as_posix(),
                              result.remote_id)
            if project_update and result is not None:
                self.safe_project.update_template_file(result)
        except Exception as e:
            self.logger.error('Failed to store remote %s %s: %s',
                              file.remote_type.value, file.filename.as_posix(), e)

    async def store_remote_files(self):
        if len(self.safe_project.safe_template.files) == 0:
            self.logger.warning('No files to store, maybe you forgot to '
                                'update _tdk.files patterns in template.json?')
        for file in self.safe_project.safe_template.files.values():
            file.remote_id = None
            file.remote_type = TemplateFileType.FILE if file.is_text else TemplateFileType.ASSET
            await self._create_template_file(file=file, project_update=True)

    def create_package(self, output: pathlib.Path, force: bool):
        if output.exists() and not force:
            raise RuntimeError(f'File {output} already exists (not forced)')
        self.logger.debug('Opening ZIP file for write: %s', output.as_posix())
        with zipfile.ZipFile(output, mode='w', compression=zipfile.ZIP_DEFLATED) as pkg:
            descriptor = self.safe_project.safe_template.serialize_for_package()
            files = []
            assets = []
            for file in self.safe_project.safe_template.files.values():
                if file.is_text:
                    self.logger.info('Adding template file %s', file.filename.as_posix())
                    files.append({
                        'uuid': str(UUIDGen.generate()),
                        'content': file.content.decode(encoding=consts.DEFAULT_ENCODING),
                        'fileName': str(file.filename.as_posix()),
                    })
                else:
                    self.logger.info('Adding template asset %s',
                                     file.filename.as_posix())
                    assets.append({
                        'uuid': str(UUIDGen.generate()),
                        'contentType': file.content_type,
                        'fileName': str(file.filename.as_posix()),
                    })
                    self.logger.debug('Packaging template asset %s',
                                      file.filename.as_posix())
                    pkg.writestr(f'template/assets/{file.filename.as_posix()}',
                                 file.content)
            descriptor['files'] = files
            descriptor['assets'] = assets
            if len(files) == 0 and len(assets) == 0:
                self.logger.warning('No files or assets found in the template, maybe you forgot '
                                    'to update _tdk.files patterns in template.json?')
            timestamp = datetime.datetime.now(tz=datetime.UTC).strftime('%Y-%m-%dT%H:%M:%S.%fZ')
            descriptor['createdAt'] = timestamp
            descriptor['updatedAt'] = timestamp
            self.logger.debug('Packaging template.json file')
            pkg.writestr('template/template.json',
                         data=json.dumps(descriptor, indent=4))
        self.logger.debug('ZIP packaging done')

    # pylint: disable=too-many-locals
    def extract_package(self, zip_data: bytes, template_dir: pathlib.Path | None, force: bool):
        with tempfile.TemporaryDirectory() as tmp_dir:
            io_zip = io.BytesIO(zip_data)
            with zipfile.ZipFile(io_zip) as pkg:
                pkg.extractall(tmp_dir)
            del io_zip
            tmp_root = pathlib.Path(tmp_dir) / 'template'
            file = tmp_root / 'template.json'
            assets_dir = tmp_root / 'assets'
            self.logger.debug('Extracting template data')
            if not file.exists():
                raise RuntimeError('Malformed package: missing template.json file')
            data = json.loads(file.read_text(encoding=consts.DEFAULT_ENCODING))
            template = Template.load_local(data)
            template.tdk_config.use_default_files()
            self.logger.warning('Using default _tdk.files in template.json, you may want '
                                'to change it to include relevant files')
            self.logger.debug('Preparing template dir')
            if template_dir is None:
                template_dir = pathlib.Path.cwd() / template.id.replace(':', '_')
            if template_dir.exists():
                if not force:
                    raise RuntimeError(f'Template directory already exists: '
                                       f'{template_dir.as_posix()} (use force?)')
                shutil.rmtree(template_dir, ignore_errors=True)
            template_dir.mkdir(parents=True)
            self.logger.debug('Extracting template.json from package')
            local_template_json = template_dir / 'template.json'
            local_template_json.write_text(
                data=json.dumps(template.serialize_local_new(), indent=2),
                encoding=consts.DEFAULT_ENCODING,
            )
            self.logger.debug('Extracting README.md from package')
            local_readme = template_dir / 'README.md'
            local_readme.write_text(
                data=data['readme'].replace('\r\n', '\n'),
                encoding=consts.DEFAULT_ENCODING,
            )
            self.logger.debug('Extracting assets from package')
            for asset_file in assets_dir.rglob('*'):
                if asset_file.is_file():
                    target_asset = template_dir / asset_file.relative_to(assets_dir)
                    target_dir = target_asset.parent
                    target_dir.mkdir(parents=True, exist_ok=True)
                    target_asset.write_bytes(asset_file.read_bytes())
            self.logger.debug('Extracting files from package')
            for file_item in data.get('files', []):
                filename = file_item['fileName']
                content = file_item['content'].replace('\r\n', '\n')
                target_file = template_dir / filename
                target_dir = target_file.parent
                target_dir.mkdir(parents=True, exist_ok=True)
                target_file.write_text(data=content, encoding=consts.DEFAULT_ENCODING)
        self.logger.debug('Extracting package done')

    async def watch_project(self, callback, stop_event: asyncio.Event):
        async for changes in watchfiles.awatch(
                self.safe_project.template_dir,
                stop_event=stop_event,
        ):
            await callback(
                change for change in ((change[0], pathlib.Path(change[1])) for change in changes)
                if self.safe_project.is_template_file(
                    change[1],
                    include_descriptor=True,
                    include_readme=True,
                )
            )

    async def update_descriptor(self):
        try:
            template_exists = await self.safe_client.check_draft_exists(
                remote_uuid=self.remote_uuid,
            )
            if template_exists:
                self.logger.info('Updating existing remote document template draft %s',
                                 self.safe_project.safe_template.coordinates)
                await self.safe_client.update_template_draft(
                    template=self.safe_project.safe_template,
                    remote_uuid=self.remote_uuid,
                )
            else:
                self.logger.info('Document template draft %s does '
                                 'not exist on remote - full sync',
                                 self.safe_project.safe_template.coordinates)
                await self.store_remote(force=False)
        except WizardCommunicationError as e:
            self.logger.error('Failed to update document template draft %s: %s',
                              self.safe_project.safe_template.coordinates, e.message)
        except Exception as e:
            self.logger.error('Failed to update document template draft %s: %s',
                              self.safe_project.safe_template.coordinates, e)

    async def delete_file(self, filepath: pathlib.Path):
        if not filepath.is_file():
            self.logger.debug('%s is not a regular file - skipping',
                              filepath.as_posix())
            return
        try:
            file = self.safe_project.get_template_file(filepath=filepath)
            if file is None:
                self.logger.info('File %s not tracked currently - skipping',
                                 filepath.as_posix())
                return
            await self._delete_template_file(file=file, project_update=True)
        except Exception as e:
            self.logger.error('Failed to delete file %s: %s',
                              filepath.as_posix(), e)

    async def update_file(self, filepath: pathlib.Path):
        if not filepath.is_file():
            self.logger.debug('%s is not a regular file - skipping',
                              filepath.as_posix())
            return
        try:
            remote_file = self.safe_project.get_template_file(filepath=filepath)
            local_file = self.safe_project.load_file(filepath=filepath)
            if remote_file is not None:
                await self._update_template_file(remote_file, local_file, project_update=True)
            else:
                await self._create_template_file(file=local_file, project_update=True)
        except Exception as e:
            self.logger.error('Failed to update file %s: %s', filepath.as_posix(), e)

    async def process_changes(self, changes: list[ChangeItem], force: bool):
        self.changes_processor.clear()
        try:
            await self.changes_processor.process_changes(changes, force)
        except Exception as e:
            self.logger.error('Failed to process changes: %s', e)


class ChangesProcessor:

    def __init__(self, tdk: TDKCore):
        self.tdk: TDKCore = tdk
        self.descriptor_change: ChangeItem | None = None
        self.readme_change: ChangeItem | None = None
        self.file_changes: list[ChangeItem] = []

    def clear(self):
        self.descriptor_change = None
        self.readme_change = None
        self.file_changes = []

    def _split_changes(self, changes: list[ChangeItem]):
        for change in changes:
            if change[1] == self.tdk.safe_project.descriptor_path:
                self.descriptor_change = change
            elif change[1] == self.tdk.safe_project.used_readme:
                self.readme_change = change
            elif self.tdk.safe_project.is_template_file(change[1]):
                self.file_changes.append(change)

    async def _process_file_changes(self):
        deleted = set()
        updated = set()
        for file_change in self.file_changes:
            self.tdk.logger.debug('Processing: %s',
                                  _change(file_change, self.tdk.safe_project.template_dir))
            change_type = file_change[0]
            filepath = file_change[1]
            if change_type == watchfiles.Change.deleted and filepath not in deleted:
                self.tdk.logger.debug('Scheduling delete operation')
                deleted.add(filepath)
                await self.tdk.delete_file(filepath)
            elif filepath not in updated:
                self.tdk.logger.debug('Scheduling update operation')
                updated.add(filepath)
                await self.tdk.update_file(filepath)

    async def _reload_descriptor(self, force: bool) -> bool:
        if self.descriptor_change is None:
            return False
        if self.descriptor_change[0] == watchfiles.Change.deleted:
            raise RuntimeError(f'Deleted {self.tdk.safe_project.descriptor_path} ... the end')
        self.tdk.logger.debug('Reloading %s file', TemplateProject.TEMPLATE_FILE)
        previous_id = self.tdk.safe_project.safe_template.coordinates
        self.tdk.safe_project.load_descriptor()
        self.tdk.safe_project.load_readme()
        new_id = self.tdk.safe_project.safe_template.coordinates
        if new_id != previous_id:
            self.tdk.logger.warning('Template ID changed from %s to %s',
                                    previous_id, new_id)
            self.tdk.safe_project.load()
            await self.tdk.store_remote(force=force)
            self.tdk.logger.info('Template fully reloaded... waiting for new changes')
            return True
        return False

    async def _reload_readme(self) -> bool:
        if self.readme_change is None:
            return False
        if self.readme_change[0] == watchfiles.Change.deleted:
            raise RuntimeError(f'Deleted used README file {self.tdk.safe_project.used_readme}')
        self.tdk.logger.debug('Reloading README file')
        self.tdk.safe_project.load_readme()
        return True

    async def _update_descriptor(self):
        if self.readme_change is not None or self.descriptor_change is not None:
            self.tdk.logger.debug('Updating template descriptor (metadata)')
            await self.tdk.update_descriptor()
            self.tdk.safe_project.template = self.tdk.safe_template

    async def process_changes(self, changes: list[ChangeItem], force: bool):
        self._split_changes(changes)
        full_reload = await self._reload_descriptor(force)
        if not full_reload:
            await self._reload_readme()
            await self._update_descriptor()
            await self._process_file_changes()
        self.tdk.logger.info('All changes processed... waiting for new changes')



import re

from . import consts
from .model import Format, PackageFilter, Step


class ValidationError(BaseException):
    def __init__(self, field_name: str, message: str):
        self.field_name = field_name
        self.message = message


def _validate_required(field_name: str, value) -> list[ValidationError]:
    if value is None:
        return [ValidationError(
            field_name=field_name,
            message='Missing but it is required',
        )]
    return []


def _validate_non_empty(field_name: str, value) -> list[ValidationError]:
    if value is not None and len(value.strip()) == 0:
        return [ValidationError(
            field_name=field_name,
            message='Cannot be empty or only-whitespace',
        )]
    return []


def _validate_content_type(field_name: str, value) -> list[ValidationError]:
    if value is not None and re.match(consts.REGEX_MIME_TYPE, value) is None:
        return [ValidationError(
            field_name=field_name,
            message='Content type should be valid IANA media type',
        )]
    return []


def _validate_extension(field_name: str, value) -> list[ValidationError]:
    if value is not None and re.match(consts.REGEX_ORGANIZATION_ID, value) is None:
        return [ValidationError(
            field_name=field_name,
            message='File extension should contain only letters, numbers, '
                    'and periods (inside-only)',
        )]
    return []


def _validate_organization_id(field_name: str, value) -> list[ValidationError]:
    if value is not None and re.match(consts.REGEX_ORGANIZATION_ID, value) is None:
        return [ValidationError(
            field_name=field_name,
            message='Organization ID may contain only letters, numbers, '
                    'periods, dashes, and underscores',
        )]
    return []


def _validate_template_id(field_name: str, value) -> list[ValidationError]:
    if value is not None and re.match(consts.REGEX_TEMPLATE_ID, value) is None:
        return [ValidationError(
            field_name=field_name,
            message='Template ID may contain only letters, numbers, '
                    'periods, dashes, and underscores',
        )]
    return []


def _validate_km_id(field_name: str, value) -> list[ValidationError]:
    if value is not None and re.match(consts.REGEX_KM_ID, value) is None:
        return [ValidationError(
            field_name=field_name,
            message='KM ID may contain only letters, numbers, '
                    'periods, dashes, and underscores',
        )]
    return []


def _validate_version(field_name: str, value) -> list[ValidationError]:
    if value is not None and re.match(consts.REGEX_SEMVER, value) is None:
        return [ValidationError(
            field_name=field_name,
            message='Version must be in semver format <NUM>.<NUM>.<NUM>',
        )]
    return []


def _validate_natural(field_name: str, value) -> list[ValidationError]:
    if value is not None and (not isinstance(value, int) or value < 1):
        return [ValidationError(
            field_name=field_name,
            message='It must be positive integer',
        )]
    return []


def _validate_metamodel_version(field_name: str, value) -> list[ValidationError]:
    if isinstance(value, int) and value > 0:
        return []
    if isinstance(value, str) and '.' in value:
        parts = value.split('.')
        if len(parts) != 2:
            return [ValidationError(
                field_name=field_name,
                message='It must be in format <MAJOR>.<MINOR>',
            )]
        try:
            major = int(parts[0])
            minor = int(parts[1])
            if major < 1 or minor < 0:
                return [ValidationError(
                    field_name=field_name,
                    message='It must be in format <MAJOR>.<MINOR> '
                            'with MAJOR >= 1 and MINOR >= 0',
                )]
            return []
        except ValueError:
            return [ValidationError(
                field_name=field_name,
                message='It must be in format <MAJOR>.<MINOR> '
                        'with MAJOR >= 1 and MINOR >= 0',
            )]
    return [ValidationError(
        field_name=field_name,
        message='It must be a positive integer or in format <MAJOR>.<MINOR>',
    )]


def _validate_package_id(field_name: str, value: str) -> list[ValidationError]:
    res = []
    if value is None:
        return res
    if not isinstance(value, str):
        return [ValidationError(
            field_name=field_name,
            message='Package ID is not a string',
        )]
    parts = value.split(':')
    if len(parts) != 3:
        res.append(ValidationError(
            field_name=field_name,
            message=f'Package ID is not valid (only {len(parts)} parts)',
        ))
    if re.match(consts.REGEX_ORGANIZATION_ID, parts[0]) is None:
        res.append(ValidationError(
            field_name=field_name,
            message='Package ID contains invalid organization id',
        ))
    if re.match(consts.REGEX_KM_ID, parts[1]) is None:
        res.append(ValidationError(
            field_name=field_name,
            message='Package ID contains invalid KM id',
        ))
    if re.match(consts.REGEX_SEMVER, parts[2]) is None:
        res.append(ValidationError(
            field_name=field_name,
            message='Package ID contains invalid version',
        ))
    return res


def _validate_jinja_options(field_name: str, value: dict[str, str]) -> list[ValidationError]:
    res = []
    if value is None:
        return res
    for k in ('template', 'content-type', 'extension'):
        if k not in value:
            res.append(ValidationError(
                field_name=field_name,
                message='Jinja option cannot be left out',
            ))
        elif value[k] is None or not isinstance(value[k], str) or len(value[k]) == 0:
            res.append(ValidationError(
                field_name=field_name,
                message='Jinja option cannot be empty',
            ))
    if 'content-type' in value:
        res.extend(_validate_content_type(
            field_name=f'{field_name}.content-type',
            value=value['content-type'],
        ))
    return res


class GenericValidator:

    def __init__(self, rules):
        self.rules = rules

    def validate_field(self, entity, field_name: str):
        for validator in self.rules.get(field_name, []):
            err = validator(field_name, getattr(entity, field_name))
            if len(err) != 0:
                raise err[0]

    def validate(self, entity, field_name_prefix: str = ''):
        for field_name, validators in self.rules.items():
            if field_name.startswith('__'):
                continue
            for validator in validators:
                err = validator(field_name_prefix + field_name, getattr(entity, field_name))
                if len(err) != 0:
                    raise err[0]
        if '__all' in self.rules:
            err = self.rules['__all'](field_name_prefix, entity)
            if len(err) != 0:
                raise err[0]

    def collect_errors(self, entity, field_name_prefix: str = '') -> list[ValidationError]:
        result = []
        for field_name, validators in self.rules.items():
            if field_name.startswith('__'):
                continue
            for validator in validators:
                result.extend(validator(
                    field_name=f'{field_name_prefix}{field_name}',
                    value=getattr(entity, field_name),
                ))
        if '__all' in self.rules:
            result.extend(self.rules['__all'](field_name_prefix, entity))
        return result


PackageFilterValidator = GenericValidator({
    'organization_id': [_validate_organization_id],
    'km_id': [_validate_km_id],
    'min_version': [_validate_version],
    'max_version': [_validate_version],
})


def _validate_package_filters(field_name: str, value: list[PackageFilter]) -> list[ValidationError]:
    res = []
    for v in value:
        res.extend(PackageFilterValidator.collect_errors(v, field_name_prefix=f'{field_name}.'))
    return res


def _validate_step(field_name_prefix: str, value: Step) -> list[ValidationError]:
    if value.name == 'jinja':
        return _validate_jinja_options(f'{field_name_prefix}options', value.options)
    return []


StepValidator = GenericValidator({
    'name': [_validate_non_empty],
    'options': [],
    '__all': _validate_step,
})


def _validate_steps(field_name: str, value: list[Step]) -> list[ValidationError]:
    res = []
    for v in value:
        res.extend(StepValidator.collect_errors(v, field_name_prefix=f'{field_name}.'))
    return res


FormatValidator = GenericValidator({
    'uuid': [_validate_required, _validate_non_empty],
    'name': [_validate_required, _validate_non_empty],
    'icon': [_validate_required, _validate_non_empty],
    'steps': [_validate_required, _validate_steps],
})


def _validate_formats(field_name: str, value: list[Format]) -> list[ValidationError]:
    res = []
    uuids = set()
    for v in value:
        if v.uuid in uuids:
            res.append(ValidationError(field_name, f'Duplicate format UUID {v.uuid}'))
        uuids.add(v.uuid)
        res.extend(FormatValidator.collect_errors(v, field_name_prefix=f'{field_name}.'))
    return res


TemplateValidator = GenericValidator({
    'template_id': [_validate_required, _validate_template_id],
    'organization_id': [_validate_required, _validate_organization_id],
    'version': [_validate_required, _validate_version],
    'name': [_validate_required, _validate_non_empty],
    'description': [_validate_required, _validate_non_empty],
    'readme': [_validate_required, _validate_non_empty],
    'license': [_validate_required, _validate_non_empty],
    'metamodel_version': [_validate_metamodel_version, _validate_required],
    'allowed_packages': [_validate_package_filters],
    'formats': [_validate_required, _validate_formats],
})



import collections
import enum
import json
import logging
import mimetypes
import pathlib
import typing

import pathspec

from . import consts


mimetypes.init()


class TemplateFileType(enum.Enum):
    ASSET = 'asset'
    FILE = 'file'


class PackageFilter:

    def __init__(self, *, organization_id: str | None = None, km_id: str | None = None,
                 min_version: str | None = None, max_version: str | None = None):
        self.organization_id = organization_id
        self.km_id = km_id
        self.min_version = min_version
        self.max_version = max_version

    @classmethod
    def load(cls, data):
        return PackageFilter(
            organization_id=data.get('orgId', None),
            km_id=data.get('kmId', None),
            min_version=data.get('minVersion', None),
            max_version=data.get('maxVersion', None),
        )

    def serialize(self):
        return {
            'orgId': self.organization_id,
            'kmId': self.km_id,
            'minVersion': self.min_version,
            'maxVersion': self.max_version,
        }


class Step:

    def __init__(self, *, name: str | None = None,
                 options: dict[str, str] | None = None):
        self.name = name
        self.options = options or {}

    @classmethod
    def load(cls, data):
        return Step(
            name=data.get('name', None),
            options=data.get('options', None),
        )

    def serialize(self):
        return {
            'name': self.name,
            'options': self.options,
        }


class Format:

    DEFAULT_ICON = 'fas fa-file'

    def __init__(self, *, uuid: str | None = None, name: str | None = None,
                 icon: str | None = None):
        self.uuid = uuid
        self.name = name
        self.icon: str = icon or self.DEFAULT_ICON
        self.steps: list[Step] = []

    @classmethod
    def load(cls, data):
        format_spec = Format(
            uuid=data.get('uuid', None),
            name=data.get('name', None),
            icon=data.get('icon', None),
        )
        for s_data in data.get('steps', []):
            format_spec.steps.append(Step.load(s_data))
        return format_spec

    def serialize(self):
        return {
            'uuid': self.uuid,
            'name': self.name,
            'icon': self.icon,
            'steps': [step.serialize() for step in self.steps],
        }


class TDKConfig:

    DEFAULT_README = 'README.md'
    DEFAULT_FILES = ['*', '!.git/', '!.env']

    def __init__(self, *, version: str | None = None, readme_file: str | None = None,
                 files: list[str] | None = None):
        self.version: str = version or consts.VERSION
        readme_file_str: str = readme_file or self.DEFAULT_README
        self.readme_file: pathlib.Path = pathlib.Path(readme_file_str)
        self.files: list[str] = files or []

    @classmethod
    def load(cls, data):
        return TDKConfig(
            version=data.get('version', consts.VERSION),
            readme_file=data.get('readmeFile', cls.DEFAULT_README),
            files=data.get('files', cls.DEFAULT_FILES),
        )

    def use_default_files(self):
        self.files = self.DEFAULT_FILES

    def serialize(self):
        return {
            'version': self.version,
            'readmeFile': str(self.readme_file),
            'files': self.files,
        }


class TemplateFile:

    DEFAULT_CONTENT_TYPE = 'application/octet-stream'
    TEMPLATE_EXTENSIONS = ('.j2', '.jinja', '.jinja2', '.jnj')

    def __init__(self, *, filename: pathlib.Path, remote_uuid: str | None = None,
                 remote_id: str | None = None, remote_type: TemplateFileType | None = None,
                 content_type: str | None = None, content: bytes = b''):
        self.remote_uuid: str | None = remote_uuid
        self.remote_id = remote_id
        self.filename = filename
        self.content = content
        self.content_type: str = content_type or self.guess_type()
        self.remote_type: TemplateFileType = remote_type or self.guess_template_file_type()

    def guess_template_file_type(self):
        return TemplateFileType.FILE if self.is_text else TemplateFileType.ASSET

    def guess_type(self) -> str:
        filename = self.filename.name
        for ext in self.TEMPLATE_EXTENSIONS:
            if filename.endswith(ext):
                return 'text/jinja2'
        guessed_type = mimetypes.guess_type(filename, strict=False)
        if guessed_type is None or guessed_type[0] is None:
            return self.DEFAULT_CONTENT_TYPE
        return guessed_type[0]

    @property
    def is_text(self):
        if getattr(self, 'remote_type', None) == TemplateFileType.FILE:
            return True
        return self.content_type.startswith('text')

    @property
    def has_remote_id(self):
        return self.remote_id is not None


class Template:

    # pylint: disable-next=too-many-arguments
    def __init__(self, *, uuid=None, template_id=None, organization_id=None,
                 version=None, name=None,
                 description=None, readme=None, template_license=None,
                 metamodel_version=None, tdk_config=None, loaded_json=None):
        self.uuid: str | None = uuid
        self.template_id: str | None = template_id
        self.organization_id: str | None = organization_id
        self.version: str | None = version
        self.name: str | None = name
        self.description: str | None = description
        self.readme: str | None = readme
        self.license: str | None = template_license
        self.metamodel_version: str = metamodel_version or consts.METAMODEL_VERSION
        self.allowed_packages: list[PackageFilter] = []
        self.formats: list[Format] = []
        self.files: dict[str, TemplateFile] = {}
        self.extras: list[TemplateFile] = []
        self.tdk_config: TDKConfig = tdk_config or TDKConfig()
        self.loaded_json: collections.OrderedDict = loaded_json or collections.OrderedDict()

    @property
    def coordinates(self) -> str:
        return f'{self.organization_id}:{self.template_id}:{self.version}'

    @property
    def id(self) -> str:
        return self.coordinates

    def id_with_org(self, organization_id: str) -> str:
        return f'{organization_id}:{self.template_id}:{self.version}'

    @classmethod
    def _common_load(cls, data):
        if 'id' in data:
            composite_id = data['id']  # type: str
            if composite_id.count(':') != 2:
                raise RuntimeError(f'Invalid template ID: {composite_id}')
            _, tmp_id, version = composite_id.split(':')
        else:
            try:
                tmp_id = data['templateId']
                version = data['version']
            except KeyError as e:
                raise RuntimeError('Cannot retrieve template ID') from e
        template = Template(
            uuid=data.get('uuid', None),
            organization_id=data.get('organizationId', '~'),
            template_id=tmp_id,
            version=version,
            name=data.get('name', 'Unknown template'),
            description=data.get('description', ''),
            template_license=data.get('license', 'no-license'),
            metamodel_version=data.get('metamodelVersion', consts.METAMODEL_VERSION),
            readme=data.get('readme', ''),
        )
        for ap_data in data.get('allowedPackages', []):
            template.allowed_packages.append(PackageFilter.load(ap_data))
        for f_data in data.get('formats', []):
            template.formats.append(Format.load(f_data))
        return template

    @classmethod
    def load_local(cls, data: collections.OrderedDict):
        template = cls._common_load(data)
        if '_tdk' in data:
            template.tdk_config = TDKConfig.load(data['_tdk'])
        template.loaded_json = data
        return template

    @classmethod
    def load_remote(cls, data: dict):
        return cls._common_load(data)

    def serialize_local(self) -> collections.OrderedDict:
        self.loaded_json['templateId'] = self.template_id
        self.loaded_json['organizationId'] = self.organization_id
        self.loaded_json['version'] = self.version
        self.loaded_json['name'] = self.name
        self.loaded_json['description'] = self.description
        self.loaded_json['license'] = self.license
        self.loaded_json['metamodelVersion'] = self.metamodel_version
        self.loaded_json['allowedPackages'] = [ap.serialize() for ap in self.allowed_packages]
        self.loaded_json['formats'] = [f.serialize() for f in self.formats]
        self.loaded_json['_tdk'] = self.tdk_config.serialize()
        return self.loaded_json

    def serialize_remote(self) -> dict[str, typing.Any]:
        return {
            'uuid': self.uuid,
            'templateId': self.template_id,
            'organizationId': self.organization_id,
            'version': self.version,
            'name': self.name,
            'description': self.description,
            'license': self.license,
            'metamodelVersion': self.metamodel_version,
            'readme': self.readme,
            'allowedPackages': [ap.serialize() for ap in self.allowed_packages],
            'formats': [f.serialize() for f in self.formats],
            'phase': 'DraftDocumentTemplatePhase',
        }

    def serialize_for_package(self) -> dict[str, typing.Any]:
        return {
            'id': self.coordinates,
            'templateId': self.template_id,
            'organizationId': self.organization_id,
            'version': self.version,
            'name': self.name,
            'description': self.description,
            'license': self.license,
            'metamodelVersion': self.metamodel_version,
            'readme': self.readme,
            'allowedPackages': [ap.serialize() for ap in self.allowed_packages],
            'formats': [f.serialize() for f in self.formats],
        }

    def serialize_for_update(self) -> dict[str, typing.Any]:
        return {
            'templateId': self.template_id,
            'version': self.version,
            'name': self.name,
            'description': self.description,
            'license': self.license,
            'metamodelVersion': self.metamodel_version,
            'readme': self.readme,
            'allowedPackages': [ap.serialize() for ap in self.allowed_packages],
            'formats': [f.serialize() for f in self.formats],
            'phase': 'DraftDocumentTemplatePhase',
        }

    def serialize_for_create(self, based_on: str | None = None) -> dict[str, typing.Any]:
        return {
            'basedOn': based_on,
            'name': self.name,
            'templateId': self.template_id,
            'version': self.version,
        }

    def serialize_local_new(self) -> dict[str, typing.Any]:
        return {
            'templateId': self.template_id,
            'organizationId': self.organization_id,
            'version': self.version,
            'name': self.name,
            'description': self.description,
            'license': self.license,
            'metamodelVersion': self.metamodel_version,
            'allowedPackages': [ap.serialize() for ap in self.allowed_packages],
            'formats': [f.serialize() for f in self.formats],
            '_tdk': self.tdk_config.serialize(),
        }


def _to_ordered_dict(tuples: list[tuple[str, typing.Any]]) -> collections.OrderedDict:
    return collections.OrderedDict(tuples)


class TemplateProject:

    TEMPLATE_FILE = 'template.json'
    DEFAULT_PATTERNS = ['!**/.*', '!**/.*/', '!**/~*', '!**/~*/',
                        '!template.json', '!template.zip']

    json_decoder = json.JSONDecoder(object_pairs_hook=_to_ordered_dict)

    def __init__(self, template_dir: pathlib.Path, logger: logging.Logger):
        self.template_dir = pathlib.Path(template_dir)
        self.descriptor_path = self.template_dir / self.TEMPLATE_FILE
        self.template: Template | None = None
        self.used_readme: pathlib.Path | None = None
        self._logger = logger

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @property
    def safe_template(self) -> Template:
        if self.template is None:
            raise RuntimeError('Template is not loaded')
        return self.template

    def load_descriptor(self):
        if not self.descriptor_path.is_file():
            raise RuntimeError(f'Template file does not exist: {self.descriptor_path.as_posix()}')
        try:
            content = self.descriptor_path.read_text(encoding=consts.DEFAULT_ENCODING)
            self.template = Template.load_local(self.json_decoder.decode(content))
        except Exception as e:
            raise RuntimeError(f'Unable to load template using {self.descriptor_path}.') from e

    def load_readme(self):
        readme = self.safe_template.tdk_config.readme_file
        if readme is not None:
            try:
                self.used_readme = self.template_dir / readme
                readme_text = self.used_readme.read_text(encoding=consts.DEFAULT_ENCODING)
                self.safe_template.readme = readme_text
            except Exception as e:
                raise RuntimeWarning(f'README file "{readme}" cannot be loaded: {e}') from e

    def load_file(self, filepath: pathlib.Path) -> TemplateFile:
        try:
            if filepath.is_absolute():
                filepath = filepath.relative_to(self.template_dir)
            template_file = TemplateFile(filename=filepath)
            template_file.content = (self.template_dir / filepath).read_bytes()
            self.safe_template.files[filepath.as_posix()] = template_file
            return template_file
        except Exception as e:
            raise RuntimeWarning(f'Failed to load template file {filepath}: {e}') from e

    def load_files(self):
        self.safe_template.files.clear()
        for f in self.list_files():
            self.load_file(f)

    @property
    def files_pathspec(self) -> pathspec.PathSpec:
        patterns = self.safe_template.tdk_config.files + self.DEFAULT_PATTERNS
        return pathspec.PathSpec.from_lines('gitignore', patterns)

    def list_files(self) -> list[pathlib.Path]:
        files = (pathlib.Path(p)
                 for p in self.files_pathspec.match_tree_files(str(self.template_dir)))
        if self.used_readme is not None:
            return [p for p in files if p != self.used_readme.relative_to(self.template_dir)]
        return list(files)

    def _relative_paths_eq(self, filepath1: pathlib.Path | None,
                           filepath2: pathlib.Path | None) -> bool:
        if filepath1 is None or filepath2 is None:
            return False
        return filepath1.relative_to(self.template_dir) == filepath2.relative_to(self.template_dir)

    def is_template_file(self, filepath: pathlib.Path, include_descriptor: bool = False,
                         include_readme: bool = False):
        if include_readme and self._relative_paths_eq(filepath, self.used_readme):
            return True
        if include_descriptor and self._relative_paths_eq(filepath, self.descriptor_path):
            return True
        return self.files_pathspec.match_file(filepath.relative_to(self.template_dir))

    def load(self):
        self.load_descriptor()
        self.load_readme()
        self.load_files()

    def remove_template_file(self, filepath: pathlib.Path):
        if filepath.is_absolute():
            filepath = filepath.relative_to(self.template_dir)
        filename = filepath.as_posix()
        if filename in self.safe_template.files:
            del self.safe_template.files[filename]

    def update_template_file(self, template_file: TemplateFile):
        filename = template_file.filename.as_posix()
        self.safe_template.files[filename] = template_file

    def get_template_file(self, filepath: pathlib.Path) -> TemplateFile | None:
        if filepath.is_absolute():
            filepath = filepath.relative_to(self.template_dir)
        return self.safe_template.files.get(filepath.as_posix(), None)

    def _write_file(self, filepath: pathlib.Path, contents: bytes, force: bool):
        if filepath.exists() and not force:
            self.logger.warning('Skipping file %s (not forced)', filepath.as_posix())
            return
        try:
            filepath.parent.mkdir(parents=True, exist_ok=True)
            filepath.write_bytes(contents)
            self.logger.debug('Stored file %s', filepath.as_posix())
        except Exception as e:
            self.logger.error('Unable to write file %s: %s', filepath.as_posix(), e)

    def store_descriptor(self, force: bool):
        self._write_file(
            filepath=self.descriptor_path,
            contents=json.dumps(
                obj=self.safe_template.serialize_local(),
                indent=4,
            ).encode(encoding=consts.DEFAULT_ENCODING),
            force=force,
        )

    def store_readme(self, force: bool):
        if self.safe_template.tdk_config.readme_file is None:
            self.logger.warning('No README file specified for the template')
            return
        if self.safe_template.readme is None:
            self.logger.warning('No README content specified for the template')
            return
        self._write_file(
            filepath=self.template_dir / self.safe_template.tdk_config.readme_file,
            contents=self.safe_template.readme.encode(encoding=consts.DEFAULT_ENCODING),
            force=force,
        )

    def store_files(self, force: bool):
        for template_file in self.safe_template.files.values():
            self._write_file(
                filepath=self.template_dir / template_file.filename,
                contents=template_file.content,
                force=force,
            )

    def store(self, force: bool):
        self.logger.debug('Ensuring directory %s', self.template_dir.as_posix())
        self.template_dir.mkdir(parents=True, exist_ok=True)
        self.logger.debug('Storing %s descriptor', self.TEMPLATE_FILE)
        self.store_descriptor(force=force)
        self.logger.debug('Storing README file')
        self.store_readme(force=force)
        self.logger.debug('Storing template files')
        self.store_files(force=force)
        self.logger.debug('Storing finished')



import functools
import pathlib
import urllib.parse

import aiohttp
import aiohttp.client_exceptions

from . import consts
from .model import Template, TemplateFile, TemplateFileType


class WizardCommunicationError(RuntimeError):

    def __init__(self, reason: str, message: str):
        """
        Exception representing communication error with DSW.

        Args:
            reason (str): Cause of the error.
            message (str): Additional information about the error.
        """
        self.reason = reason
        self.message = message


def handle_client_errors(func):
    @functools.wraps(func)
    async def handled_client_call(job, *args, **kwargs):
        try:
            return await func(job, *args, **kwargs)
        except WizardCommunicationError as e:
            # Already DSWCommunicationError (re-raise)
            raise e
        except aiohttp.client_exceptions.ContentTypeError as e:
            raise WizardCommunicationError(
                reason='Unexpected response type',
                message=e.message,
            ) from e
        except aiohttp.client_exceptions.ClientResponseError as e:
            raise WizardCommunicationError(
                reason='Error response status',
                message=f'Server responded with error HTTP status {e.status}: {e.message}',
            ) from e
        except aiohttp.client_exceptions.InvalidURL as e:
            raise WizardCommunicationError(
                reason='Invalid URL',
                message=f'Provided API URL seems invalid: {e.url}',
            ) from e
        except aiohttp.client_exceptions.ClientConnectorError as e:
            raise WizardCommunicationError(
                reason='Server unreachable',
                message=f'Desired server is not reachable (errno {e.os_error.errno})',
            ) from e
        except Exception as e:
            raise WizardCommunicationError(
                reason='Communication error',
                message=f'Communication with server failed ({e})',
            ) from e
    return handled_client_call


# pylint: disable-next=too-many-public-methods
class WizardAPIClient:

    def _headers(self, extra=None):
        headers = {
            'Authorization': f'Bearer {self.token}',
            'User-Agent': f'{consts.APP}/{consts.VERSION}',
        }
        if extra is not None:
            headers.update(extra)
        return headers

    @staticmethod
    def _check_status(r: aiohttp.ClientResponse, expected_status):
        r.raise_for_status()
        if r.status != expected_status:
            raise WizardCommunicationError(
                reason='Unexpected response status',
                message=f'Server responded with unexpected HTTP status {r.status}: '
                        f'{r.reason} (expecting {expected_status})',
            )

    def __init__(self, api_url: str, api_key: str, session=None):
        """
        Exception representing communication error with DSW.

        Args:
            api_url (str): URL of DSW API for HTTP communication.
            session (aiohttp.ClientSession): Optional custom session for HTTP communication.
        """
        self.api_url = api_url
        self.client_url = api_url[:-4]
        self.token = api_key
        self.session = session or aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=False),
        )

    def template_editor_url(self, template_uuid: str) -> str:
        return f'{self.client_url}/editor/{template_uuid}'

    @property
    def templates_endpoint(self):
        return f'{self.api_url}/document-templates'

    @property
    def drafts_endpoint(self):
        return f'{self.api_url}/document-template-drafts'

    async def close(self):
        await self.session.close()

    async def safe_close(self) -> bool:
        try:
            await self.session.close()
            return True
        except Exception:
            return False

    async def _post_json(self, endpoint: str, json) -> dict:
        async with self.session.post(
            url=f'{self.api_url}{endpoint}',
            json=json,
            headers=self._headers(),
        ) as r:
            self._check_status(r, expected_status=201)
            return await r.json()

    async def _put_json(self, endpoint: str, json) -> dict:
        async with self.session.put(
            url=f'{self.api_url}{endpoint}',
            json=json,
            headers=self._headers(),
        ) as r:
            self._check_status(r, expected_status=200)
            return await r.json()

    async def _get_json(self, endpoint: str, params=None) -> dict:
        async with self.session.get(
            url=f'{self.api_url}{endpoint}',
            headers=self._headers(),
            params=params,
        ) as r:
            self._check_status(r, expected_status=200)
            return await r.json()

    async def _get_bytes(self, endpoint: str) -> bytes:
        async with self.session.get(
            url=f'{self.api_url}{endpoint}',
            headers=self._headers(),
        ) as r:
            self._check_status(r, expected_status=200)
            return await r.read()

    async def _delete(self, endpoint: str) -> bool:
        async with self.session.delete(
            url=f'{self.api_url}{endpoint}',
            headers=self._headers(),
        ) as r:
            return r.status == 204

    @handle_client_errors
    async def login(self, email: str, password: str) -> str | None:
        req = {'email': email, 'password': password}
        body = await self._post_json('/tokens', json=req)
        token_value = body.get('token')
        if not isinstance(token_value, str):
            raise WizardCommunicationError(
                reason='Invalid response',
                message='Server did not return a valid token',
            )
        self.token = token_value
        return self.token

    @handle_client_errors
    async def get_current_user(self) -> dict:
        return await self._get_json('/users/current')

    @handle_client_errors
    async def check_template_exists(self, remote_id: str) -> bool:
        async with self.session.get(
                url=f'{self.templates_endpoint}/{remote_id}',
                headers=self._headers(),
        ) as r:
            if r.status == 404:
                return False
            self._check_status(r, expected_status=200)
            return True

    @handle_client_errors
    async def check_draft_exists(self, remote_uuid: str) -> bool:
        async with self.session.get(
                url=f'{self.drafts_endpoint}/{remote_uuid}',
                headers=self._headers(),
        ) as r:
            if r.status == 404:
                return False
            self._check_status(r, expected_status=200)
            return True

    @handle_client_errors
    async def get_templates(self) -> list[Template]:
        body = await self._get_json('/document-templates/all')
        return list(map(_load_remote_template, body))

    @handle_client_errors
    async def get_drafts(self) -> list[Template]:
        body = await self._get_json('/document-template-drafts?size=10000')
        drafts = body.get('_embedded', {}).get('documentTemplateDrafts', [])
        return list(map(_load_remote_template, drafts))

    @handle_client_errors
    async def get_template_bundle(self, remote_uuid: str) -> bytes:
        remote_file_descriptor = await self._get_json(
            endpoint=f'/document-templates/{remote_uuid}/bundle'
                     f'?Authorization=Bearer%20{self.token}',
        )
        s3_url = remote_file_descriptor['url']
        async with self.session.get(url=s3_url) as r:
            self._check_status(r, expected_status=200)
            return await r.content.read()

    async def _find_by_id(self, template_id: str, endpoint: str, key: str) -> str | None:
        org_id, t_id, version = template_id.split(':')
        body = await self._get_json(
            endpoint=endpoint,
            params={
                'size': 10000,
                'page': 0,
                'q': t_id,
            },
        )
        drafts = body.get('_embedded', {}).get(key, [])
        partial_match = None
        for draft in drafts:
            v_match = draft.get('version') == version
            o_match = draft.get('organizationId') == org_id
            t_match = draft.get('templateId') == t_id
            if o_match and t_match and v_match:
                return draft.get('uuid')
            if t_match and v_match:
                partial_match = draft.get('uuid')
        return partial_match

    @handle_client_errors
    async def find_template_uuid_by_id(self, template_id: str) -> str | None:
        return await self._find_by_id(
            template_id=template_id,
            endpoint='/document-templates',
            key='documentTemplates',
        )

    @handle_client_errors
    async def find_template_draft_uuid_by_id(self, template_id: str) -> str | None:
        return await self._find_by_id(
            template_id=template_id,
            endpoint='/document-template-drafts',
            key='documentTemplateDrafts',
        )

    @handle_client_errors
    async def get_template_draft(self, remote_uuid: str) -> Template:
        body = await self._get_json(f'/document-template-drafts/{remote_uuid}')
        return _load_remote_template(body)

    @handle_client_errors
    async def get_template_draft_files(self, remote_uuid: str) -> list[TemplateFile]:
        body = await self._get_json(f'/document-template-drafts/{remote_uuid}/files')
        result = []
        for file_body in body:
            file_id = file_body['uuid']
            file = await self.get_template_draft_file(remote_uuid, file_id)
            result.append(file)
        return result

    @handle_client_errors
    async def get_template_draft_file(self, remote_uuid: str, file_id: str) -> TemplateFile:
        body = await self._get_json(f'/document-template-drafts/{remote_uuid}/files/{file_id}')
        return _load_remote_file(body)

    @handle_client_errors
    async def get_template_draft_assets(self, remote_uuid: str) -> list[TemplateFile]:
        body = await self._get_json(f'/document-template-drafts/{remote_uuid}/assets')
        result = []
        for file_body in body:
            asset_id = file_body['uuid']
            template_asset = await self.get_template_draft_asset(remote_uuid, asset_id)
            result.append(template_asset)
        return result

    @handle_client_errors
    async def get_template_draft_asset(self, remote_uuid: str, asset_id: str) -> TemplateFile:
        body = await self._get_json(f'/document-template-drafts/{remote_uuid}'
                                    f'/assets/{asset_id}')
        content = await self._get_bytes(f'/document-template-drafts/{remote_uuid}'
                                        f'/assets/{asset_id}/content')
        return _load_remote_asset(body, content)

    @handle_client_errors
    async def create_new_template_draft(self, template: Template, remote_id: str) -> Template:
        result = await self._post_json(
            endpoint='/document-template-drafts',
            json=template.serialize_for_create(),
        )
        remote_uuid = result.get('uuid', '')
        new_draft = await self.get_template_draft(remote_uuid)
        parts = remote_id.split(':')
        if new_draft.template_id != parts[1] or new_draft.version != parts[2]:
            raise WizardCommunicationError(
                reason='Invalid response',
                message='Server did not return a draft with expected template ID and version',
            )
        body = await self._put_json(
            endpoint=f'/document-template-drafts/{remote_uuid}',
            json=template.serialize_for_update(),
        )
        return _load_remote_template(body)

    @handle_client_errors
    async def update_template_draft(self, template: Template, remote_uuid: str) -> Template:
        body = await self._put_json(
            endpoint=f'/document-template-drafts/{remote_uuid}',
            json=template.serialize_for_update(),
        )
        return _load_remote_template(body)

    @handle_client_errors
    async def post_template_draft_file(self, remote_uuid: str, file: TemplateFile):
        data = await self._post_json(
            endpoint=f'/document-template-drafts/{remote_uuid}/files',
            json={
                'fileName': file.filename.as_posix(),
                'content': file.content.decode(consts.DEFAULT_ENCODING),
            },
        )
        return _load_remote_file(data)

    @handle_client_errors
    async def put_template_draft_file_content(self, remote_uuid: str, file: TemplateFile):
        self.session.headers.update(self._headers())
        async with self.session.put(
                f'{self.api_url}/document-template-drafts/{remote_uuid}'
                f'/files/{file.remote_id}/content',
                data=file.content,
                headers={'Content-Type': 'text/plain;charset=UTF-8'},
        ) as r:
            self._check_status(r, expected_status=200)
            body = await r.json()
            return _load_remote_file(body)

    @handle_client_errors
    async def post_template_draft_asset(self, remote_uuid: str, file: TemplateFile):
        data = aiohttp.FormData()
        data.add_field(
            name='file',
            value=file.content,
            filename=file.filename.as_posix(),
            content_type=file.content_type,
        )
        data.add_field(
            name='fileName',
            value=file.filename.as_posix(),
        )
        async with self.session.post(
                f'{self.api_url}/document-template-drafts/{remote_uuid}/assets',
                data=data,
                headers=self._headers(),
        ) as r:
            self._check_status(r, expected_status=201)
            body = await r.json()
            return _load_remote_asset(body, file.content)

    @handle_client_errors
    async def put_template_draft_asset_content(self, remote_uuid: str, file: TemplateFile):
        data = aiohttp.FormData()
        data.add_field(
            name='file',
            value=file.content,
            filename=file.filename.as_posix(),
            content_type=file.content_type,
        )
        data.add_field(
            name='fileName',
            value=file.filename.as_posix(),
        )
        async with self.session.put(
                f'{self.api_url}/document-template-drafts/{remote_uuid}'
                f'/assets/{file.remote_id}/content',
                data=data,
                headers=self._headers(),
        ) as r:
            self._check_status(r, expected_status=200)
            body = await r.json()
            return _load_remote_asset(body, file.content)

    @handle_client_errors
    async def delete_template_draft(self, remote_uuid: str) -> bool:
        return await self._delete(f'/document-template-drafts/{remote_uuid}')

    @handle_client_errors
    async def delete_template_draft_file(self, remote_uuid: str, file_uuid: str) -> bool:
        if file_uuid is None:
            raise RuntimeWarning('Tried to delete file without ID (None)')
        return await self._delete(f'/document-template-drafts/{remote_uuid}/files/{file_uuid}')

    @handle_client_errors
    async def delete_template_draft_asset(self, remote_uuid: str, asset_uuid: str) -> bool:
        if asset_uuid is None:
            raise RuntimeWarning('Tried to delete asset without ID (None)')
        return await self._delete(f'/document-template-drafts/{remote_uuid}/assets/{asset_uuid}')

    @handle_client_errors
    async def get_api_version(self) -> tuple[str, str | None]:
        body = await self._get_json('/')
        version = body.get('version')
        metamodel_version = None
        for item in body.get('metamodelVersions', []):
            if item.get('name', '') == 'Document Template':
                metamodel_version = item.get('version')
        if version is None:
            raise WizardCommunicationError(
                reason='Invalid response',
                message='Server did not return valid API version information (incompatible TDK?)',
            )
        return version, metamodel_version

    @handle_client_errors
    async def get_organization_id(self) -> str:
        body = await self._get_json('/configs/bootstrap')
        return body['organization']['organizationId']


def _load_remote_file(data: dict) -> TemplateFile:
    content: str = data.get('content', '')
    filename: str = str(data.get('fileName', ''))
    return TemplateFile(
        remote_id=data.get('uuid'),
        remote_type=TemplateFileType.FILE,
        filename=pathlib.Path(urllib.parse.unquote(filename)),
        content=content.encode(encoding=consts.DEFAULT_ENCODING),
    )


def _load_remote_asset(data: dict, content: bytes) -> TemplateFile:
    filename = str(data.get('fileName', ''))
    return TemplateFile(
        remote_id=data.get('uuid'),
        remote_type=TemplateFileType.ASSET,
        filename=pathlib.Path(urllib.parse.unquote(filename)),
        content_type=data.get('contentType'),
        content=content,
    )


def _load_remote_template(data: dict) -> Template:
    return Template.load_remote(data)



import pathlib
import re

import pathspec


APP = 'dsw-tdk'
VERSION = '4.28.4'
METAMODEL_VERSION_MAJOR = 17
METAMODEL_VERSION_MINOR = 1
METAMODEL_VERSION = f'{METAMODEL_VERSION_MAJOR}.{METAMODEL_VERSION_MINOR}'

REGEX_SEMVER = re.compile(r'^[0-9]+\.[0-9]+\.[0-9]+$')
REGEX_WIZARD_ID = re.compile(r'^[a-zA-Z0-9-_.]+$')
REGEX_ORGANIZATION_ID = REGEX_WIZARD_ID
REGEX_TEMPLATE_ID = REGEX_WIZARD_ID
REGEX_KM_ID = REGEX_WIZARD_ID
REGEX_MIME_TYPE = re.compile(r'^(?![-])(?!.*[-]$)[-\w.]+/[-\w.]+$')

DEFAULT_LIST_FORMAT = '{template.id:<50} {template.name:<30} [{template.uuid}]'
DEFAULT_ENCODING = 'utf-8'
DEFAULT_README = pathlib.Path('README.md')

TEMPLATE_FILE = 'template.json'
PathspecFactory = pathspec.patterns.GitWildMatchPattern



import configparser
import dataclasses
import os
import pathlib

import dotenv

from . import consts


def _rectify_api_url(api_url: str | None) -> str:
    if not api_url or not isinstance(api_url, str):
        return ''
    return api_url.rstrip('/')


def _rectify_api_key(api_key: str | None) -> str:
    if not api_key or not isinstance(api_key, str):
        return ''
    return api_key.strip()


@dataclasses.dataclass
class TDKWizardEnv:
    api_url: str
    api_key: str

    def rectify(self):
        self.api_url = _rectify_api_url(self.api_url)
        self.api_key = _rectify_api_key(self.api_key)


class TDKConfig:
    LOCAL_CONFIG = '_local'
    HOME_CONFIG = pathlib.Path.home() / '.dsw-tdk' / 'config.cfg'

    def __init__(self):
        self.shared_envs = {}  # type: dict[str, TDKWizardEnv]
        self.local_env = TDKWizardEnv(
            api_url='',
            api_key='',
        )
        self.current_env_name = self.LOCAL_CONFIG
        self.default_env_name = None  # type: str | None

    def load_dotenv(self, path: pathlib.Path):
        if path.exists():
            dotenv.load_dotenv(path)
        api_url = os.getenv('DSW_API_URL', '')
        api_key = os.getenv('DSW_API_KEY', '')

        if not api_url or not api_key:
            return

        self.current_env_name = self.LOCAL_CONFIG
        self.local_env = TDKWizardEnv(
            api_url=api_url,
            api_key=api_key,
        )
        self.local_env.rectify()

    def load_home_config(self):
        config_path = self.HOME_CONFIG
        if not config_path.exists():
            return

        config = configparser.ConfigParser()
        config.read(config_path)
        for section in config.sections():
            if section == 'default':
                self.default_env_name = config.get(section, 'env', fallback=None)
            elif section.startswith('env:'):
                env_name = section[4:]
                if env_name not in self.shared_envs:
                    self.shared_envs[env_name] = TDKWizardEnv(
                        api_url=config.get(section, 'api_url', fallback=''),
                        api_key=config.get(section, 'api_key', fallback=''),
                    )
                    self.shared_envs[env_name].rectify()
        if self.default_env_name is not None:
            self.current_env_name = self.default_env_name

    @property
    def env(self) -> TDKWizardEnv:
        if self.current_env_name == self.LOCAL_CONFIG:
            return self.local_env
        return self.shared_envs[self.current_env_name]

    @property
    def has_api_url(self) -> bool:
        return self.env.api_url != ''

    def set_api_url(self, api_url: str):
        self.env.api_url = _rectify_api_url(api_url)

    @property
    def has_api_key(self) -> bool:
        return self.env.api_key != ''

    def set_api_key(self, api_key: str):
        self.env.api_key = _rectify_api_key(api_key)

    def use_local_env(self):
        self.current_env_name = self.LOCAL_CONFIG

    def switch_current_env(self, env_name: str | None):
        if env_name not in self.shared_envs:
            raise ValueError(f'Environment "{env_name}" does not exist')
        self.current_env_name = env_name

    @property
    def is_default_env(self) -> bool:
        return self.current_env_name == self.default_env_name

    @property
    def env_names(self) -> list[str]:
        return sorted(env_name for env_name in self.shared_envs)

    def add_shared_env(self, name: str, api_url: str, api_key: str):
        if name == self.LOCAL_CONFIG:
            raise ValueError(f'Environment name "{self.LOCAL_CONFIG}" is reserved')
        if name in self.shared_envs:
            raise ValueError(f'Environment "{name}" already exists')
        self.shared_envs[name] = TDKWizardEnv(
            api_url=api_url,
            api_key=api_key,
        )

    def persist(self, force: bool):
        output = self.HOME_CONFIG
        if output.exists() and not force:
            raise FileExistsError(f'File {output.as_posix()} already exists (not forced)')
        if not output.parent.exists():
            output.parent.mkdir(parents=True, exist_ok=True)

        config = configparser.ConfigParser()
        config.add_section('default')
        if self.default_env_name:
            config.set('default', 'env', self.default_env_name)

        for env_name, env in self.shared_envs.items():
            section_name = f'env:{env_name}'
            config.add_section(section_name)
            config.set(section_name, 'api_url', env.api_url)
            config.set(section_name, 'api_key', env.api_key)

        with output.open(mode='w', encoding=consts.DEFAULT_ENCODING) as config_file:
            config.write(config_file)


CONFIG = TDKConfig()



# pylint: disable=too-many-positional-arguments
import asyncio
import datetime
import logging
import mimetypes
import pathlib
import signal
import sys

import click
import humanize
import slugify
import watchfiles

from . import consts
from .api_client import WizardCommunicationError
from .config import CONFIG
from .core import TDKCore, TDKProcessingError
from .model import Template
from .utils import FormatSpec, TemplateBuilder, create_dot_env, safe_utf8
from .validation import ValidationError


CURRENT_DIR = pathlib.Path.cwd()
DIR_TYPE = click.Path(exists=True, dir_okay=True, file_okay=False, resolve_path=True,
                      readable=True, writable=True)
FILE_READ_TYPE = click.Path(exists=True, dir_okay=False, file_okay=True, resolve_path=True,
                            readable=True)
NEW_DIR_TYPE = click.Path(dir_okay=True, file_okay=False, resolve_path=True,
                          readable=True, writable=True)


def _now() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.UTC)


class ClickPrinter:

    CHANGE_SIGNS = {
        watchfiles.Change.added: click.style('+', fg='green'),
        watchfiles.Change.modified: click.style('*', fg='yellow'),
        watchfiles.Change.deleted: click.style('-', fg='red'),
    }

    @staticmethod
    def error(message: str, **kwargs):
        click.secho(message=message, err=True, fg='red', **kwargs)

    @staticmethod
    def warning(message: str, **kwargs):
        click.secho('WARNING', fg='yellow', bold=True, nl=False, **kwargs)
        click.echo(f': {message}')

    @staticmethod
    def success(message: str):
        click.secho('SUCCESS', fg='green', bold=True, nl=False)
        click.echo(f': {message}')

    @staticmethod
    def failure(message: str):
        click.secho('FAILURE', fg='red', bold=True, nl=False)
        click.echo(f': {message}')

    @staticmethod
    def watch(message: str):
        click.secho('WATCH', fg='blue', bold=True, nl=False)
        click.echo(f': {message}')

    @classmethod
    def watch_change(cls, change_type: watchfiles.Change, filepath: pathlib.Path,
                     root: pathlib.Path):
        timestamp = _now().isoformat(timespec='milliseconds')
        sign = cls.CHANGE_SIGNS[change_type]
        click.secho('WATCH', fg='blue', bold=True, nl=False)
        click.echo(f'@{timestamp} {sign} {filepath.relative_to(root)}')


def prompt_fill(text: str, obj, attr, **kwargs):
    while True:
        try:
            value = safe_utf8(click.prompt(text, **kwargs).strip())
            setattr(obj, attr, value)
            break
        except ValidationError as e:
            ClickPrinter.error(e.message)


def print_template_info(template: Template):
    click.echo(f'Template ID: {template.coordinates}')
    click.echo(f'Name:        {template.name}')
    click.echo(f'License:     {template.license}')
    click.echo(f'Description: {template.description}')
    click.echo('Formats:')
    for format_spec in template.formats:
        click.echo(f' - {format_spec.name}')
    click.echo('Files:')
    for template_file in template.files.values():
        filesize = humanize.naturalsize(len(template_file.content))
        click.echo(f' - {template_file.filename.as_posix()} [{filesize}]')


def ensure_api_config(api_url: str | None, api_key: str | None):
    if api_url is not None:
        CONFIG.use_local_env()
        CONFIG.set_api_url(api_url)
    if not CONFIG.has_api_url:
        CONFIG.set_api_url(
            api_url=click.prompt('API URL'),
        )
    if api_key is not None:
        CONFIG.use_local_env()
        CONFIG.set_api_key(api_key)
    if not CONFIG.has_api_key:
        CONFIG.set_api_key(
            api_key=click.prompt('API Key', hide_input=True),
        )
    if not CONFIG.has_api_url or not CONFIG.has_api_key:
        ClickPrinter.error('API URL and API Key are required to proceed.')
        sys.exit(1)


class ClickLogger(logging.Logger):

    NAME = 'DSW-TDK-CLI'
    LEVEL_STYLES = {
        logging.CRITICAL: lambda x: click.style(x, fg='red', bold=True),
        logging.ERROR: lambda x: click.style(x, fg='red'),
        logging.WARNING: lambda x: click.style(x, fg='yellow'),
        logging.INFO: lambda x: click.style(x, fg='cyan'),
        logging.DEBUG: lambda x: click.style(x, fg='magenta'),
    }
    LEVELS = [
        logging.getLevelName(logging.CRITICAL),
        logging.getLevelName(logging.ERROR),
        logging.getLevelName(logging.WARNING),
        logging.getLevelName(logging.INFO),
        logging.getLevelName(logging.DEBUG),
    ]

    def __init__(self, show_timestamp: bool = False, show_level: bool = True, colors: bool = True):
        super().__init__(name=self.NAME)
        self.show_timestamp = show_timestamp
        self.show_level = show_level
        self.colors = colors
        self.muted = False

    def _format_level(self, level, justify=False):
        name = logging.getLevelName(level)  # type: str
        if justify:
            name = name.ljust(8, ' ')
        if self.colors and level in self.LEVEL_STYLES:
            name = self.LEVEL_STYLES[level](name)
        return name

    def _print_message(self, level, message):
        if self.show_timestamp:
            timestamp = _now().isoformat(timespec='milliseconds')
            click.echo(timestamp + ' | ', nl=False)
        if self.show_level:
            sep = ' | ' if self.show_timestamp else ': '
            click.echo(self._format_level(level, justify=self.show_timestamp) + sep, nl=False)
        click.echo(message)

    # pylint: disable-next=unused-argument
    def _log(self, level, msg, args, *other, **kwargs):
        if not self.muted and isinstance(msg, str):
            self._print_message(level, msg % args)

    @staticmethod
    def default():
        logger = ClickLogger()
        logger.setLevel('INFO')
        return logger


class AliasedGroup(click.Group):

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        if len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        return ctx.fail(f'Too many matches: {", ".join(sorted(matches))}')


class CLIContext:

    def __init__(self):
        self.logger = ClickLogger.default()
        self.dot_env_file = None

    def debug_mode(self):
        self.logger.show_timestamp = True
        self.logger.setLevel(level=logging.DEBUG)

    def quiet_mode(self):
        self.logger.muted = True


def interact_formats() -> dict[str, FormatSpec]:
    add_format = click.confirm('Do you want to add a format?', default=True)
    formats: dict[str, FormatSpec] = {}
    while add_format:
        format_spec = FormatSpec()
        prompt_fill('Format name', obj=format_spec, attr='name', default='HTML')
        if format_spec.name not in formats or click.confirm(
                'There is already a format with this name. Do you want to change it?',
        ):
            prompt_fill('File extension', obj=format_spec, attr='file_extension',
                        default=format_spec.name.lower() if ' ' not in format_spec.name else None)
            prompt_fill('Content type', obj=format_spec, attr='content_type',
                        default=mimetypes.types_map.get(f'.{format_spec.file_extension}', None))
            t_path = pathlib.Path('src') / f'template.{format_spec.file_extension}.j2'
            prompt_fill(
                text='Jinja2 filename',
                obj=format_spec,
                attr='filename',
                default=str(t_path),
            )
            formats[format_spec.name] = format_spec
        click.echo('=' * 60)
        add_format = click.confirm('Do you want to add yet another format?', default=False)
    return formats


def interact_builder(builder: TemplateBuilder):
    prompt_fill('Template name', obj=builder, attr='name')
    prompt_fill('Organization ID', obj=builder, attr='organization_id')
    prompt_fill('Template ID', obj=builder, attr='template_id',
                default=slugify.slugify(builder.name))
    prompt_fill('Version', obj=builder, attr='version',
                default='0.1.0')
    prompt_fill('Description', obj=builder, attr='description',
                default='My custom template')
    prompt_fill('License', obj=builder, attr='license',
                default='CC0')
    click.echo('=' * 60)
    formats = interact_formats()
    for format_spec in formats.values():
        builder.add_format(format_spec)


def load_local(tdk: TDKCore, template_dir: pathlib.Path):
    try:
        tdk.load_local(template_dir=template_dir)
    except Exception as e:
        ClickPrinter.failure('Could not load local template')
        ClickPrinter.error(f'> {e}')
        sys.exit(1)


def dir_from_id(template_id: str) -> pathlib.Path:
    return pathlib.Path.cwd() / template_id.replace(':', '_')


@click.group(cls=AliasedGroup)
@click.option('-d', '--dot-env', default='.env', required=False,
              show_default=True, type=click.Path(file_okay=True, dir_okay=False),
              help='File path to dot-env file with environment variables.')
@click.option('-e', '--environment', default=None, required=False,
              help='Configuration environment name.')
@click.option('--no-dot-env', is_flag=True, default=False,
              help='Do not load .env file, use only environment variables.')
@click.option('--no-config', is_flag=True, default=False,
              help='Do not load shared configuration, use only environment variables.')
@click.option('-q', '--quiet', is_flag=True,
              help='Hide additional information logs.')
@click.option('--debug', is_flag=True,
              help='Enable debug logging.')
@click.version_option(version=consts.VERSION)
@click.pass_context
def main(ctx, quiet, debug, dot_env, environment, no_dot_env, no_config):
    if not no_config:
        try:
            CONFIG.load_home_config()
        except Exception as e:
            ClickPrinter.warning('Failed to load shared configuration')
            ClickPrinter.warning(f'> {e}')
    if not no_dot_env and dot_env is not None:
        dot_env_path = pathlib.Path(dot_env)
        if dot_env_path.exists():
            CONFIG.load_dotenv(path=dot_env_path)
    try:
        if environment is not None:
            CONFIG.switch_current_env(environment)
    except Exception as e:
        ClickPrinter.warning('Failed to set config environment')
        ClickPrinter.warning(f'> {e}')
    ctx.ensure_object(CLIContext)
    ctx.obj.dot_env_file = dot_env
    if quiet:
        ctx.obj.quiet_mode()
    if debug:
        ctx.obj.debug_mode()


@main.command(help='Create a new template project.', name='new')
@click.argument('TEMPLATE-DIR', type=NEW_DIR_TYPE, default=None, required=False)
@click.option('-f', '--force', is_flag=True, help='Overwrite any matching files.')
@click.pass_context
def new_template(ctx, template_dir, force):
    builder = TemplateBuilder()
    try:
        interact_builder(builder)
    except Exception:
        click.echo('')
        ClickPrinter.failure('Exited...')
        sys.exit(1)
    tdk = TDKCore(template=builder.build(), logger=ctx.obj.logger)
    template_dir = template_dir or dir_from_id(tdk.safe_template.coordinates)
    tdk.prepare_local(template_dir=template_dir)
    try:
        tdk.store_local(force=force)
        ClickPrinter.success(f'Template project created: {template_dir}')
    except Exception as e:
        ClickPrinter.failure('Could not create new template project')
        ClickPrinter.error(f'> {e}')
        sys.exit(1)


@main.command(help='Download template from Wizard.', name='get')
@click.argument('TEMPLATE-ID')
@click.argument('TEMPLATE-DIR', type=NEW_DIR_TYPE, default=None, required=False)
@click.option('-u', '--api-url', metavar='API-URL', envvar='DSW_API_URL',
              help='URL of Wizard server API.')
@click.option('-k', '--api-key', metavar='API-KEY', envvar='DSW_API_KEY',
              help='API key for Wizard instance.')
@click.option('-f', '--force', is_flag=True, help='Overwrite any existing files.')
@click.pass_context
def get_template(ctx, template_id, template_dir, api_url, api_key, force):
    ensure_api_config(api_url, api_key)
    template_dir = pathlib.Path(template_dir or dir_from_id(template_id))

    async def main_routine():
        tdk = TDKCore(logger=ctx.obj.logger)
        template_type = 'unknown'
        zip_data = None
        try:
            await tdk.init_client(
                api_url=CONFIG.env.api_url,
                api_key=CONFIG.env.api_key,
            )
            try:
                await tdk.load_remote(template_id=template_id)
                template_type = 'draft'
            except Exception:
                zip_data = await tdk.download_bundle(template_id=template_id)
                template_type = 'bundle'
            await tdk.safe_client.close()
        except WizardCommunicationError as e:
            ClickPrinter.error('Could not get template:', bold=True)
            ClickPrinter.error(f'> {e.reason}\n> {e.message}')
            await tdk.safe_client.close()
            sys.exit(1)
        await tdk.safe_client.safe_close()
        if template_type == 'draft':
            tdk.prepare_local(template_dir=template_dir)
            try:
                tdk.store_local(force=force)
                ClickPrinter.success(f'Template draft {template_id} '
                                     f'downloaded to {template_dir}')
            except Exception as e:
                ClickPrinter.failure('Could not store template locally')
                ClickPrinter.error(f'> {e}')
                await tdk.safe_client.close()
                sys.exit(1)
        elif template_type == 'bundle' and zip_data is not None:
            try:
                tdk.extract_package(zip_data=zip_data, template_dir=template_dir, force=force)
                ClickPrinter.success(f'Template {template_id} (released) '
                                     f'downloaded to {template_dir}')
            except Exception as e:
                ClickPrinter.failure('Could not store template locally')
                ClickPrinter.error(f'> {e}')
                await tdk.safe_client.close()
                sys.exit(1)
        else:
            ClickPrinter.failure(f'{template_id} is not released nor draft of a document template')
            sys.exit(1)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(main_routine())


@main.command(help='Upload template to Wizard.', name='put')
@click.argument('TEMPLATE-DIR', type=DIR_TYPE, default=CURRENT_DIR, required=False)
@click.option('-u', '--api-url', metavar='API-URL', envvar='DSW_API_URL',
              help='URL of Wizard server API.')
@click.option('-k', '--api-key', metavar='API-KEY', envvar='DSW_API_KEY',
              help='API key for Wizard instance.')
@click.option('-f', '--force', is_flag=True,
              help='Delete template if already exists.')
@click.option('-w', '--watch', is_flag=True,
              help='Enter watch mode to continually upload changes.')
@click.pass_context
def put_template(ctx, template_dir, api_url, api_key, force, watch):
    ensure_api_config(api_url, api_key)
    tdk = TDKCore(logger=ctx.obj.logger)
    stop_event = asyncio.Event()

    async def watch_callback(changes):
        changes = list(changes)
        for change in changes:
            ClickPrinter.watch_change(
                change_type=change[0],
                filepath=change[1],
                root=tdk.safe_project.template_dir,
            )
        if len(changes) > 0:
            await tdk.process_changes(changes, force=force)

    async def main_routine():
        load_local(tdk, template_dir)
        try:
            await tdk.init_client(
                api_url=CONFIG.env.api_url,
                api_key=CONFIG.env.api_key,
            )
            await tdk.store_remote(force=force)
            ClickPrinter.success(f'Template {tdk.safe_project.safe_template.coordinates} '
                                 f'uploaded to {CONFIG.env.api_url}')

            if watch:
                ClickPrinter.watch('Entering watch mode... (press Ctrl+C to abort)')
                await tdk.watch_project(watch_callback, stop_event)

            await tdk.safe_client.close()
        except TDKProcessingError as e:
            ClickPrinter.failure('Could not upload template')
            ClickPrinter.error(f'> {e.message}\n> {e.hint}')
            await tdk.safe_client.safe_close()
            sys.exit(1)
        except WizardCommunicationError as e:
            ClickPrinter.failure('Could not upload template')
            ClickPrinter.error(f'> {e.reason}\n> {e.message}')
            ClickPrinter.error('> Probably incorrect API URL, metamodel version, '
                               'or template already exists...')
            ClickPrinter.error('> Check if you are using the matching version')
            await tdk.safe_client.safe_close()
            sys.exit(1)

    # pylint: disable-next=unused-argument
    def set_stop_event(signal_num, frame):
        signal_name = signal.Signals(signal_num).name
        ClickPrinter.warning(f'Got {signal_name}, finishing... Bye!')
        stop_event.set()

    signal.signal(signal.SIGINT, set_stop_event)
    signal.signal(signal.SIGABRT, set_stop_event)

    loop = asyncio.get_event_loop()
    main_task = asyncio.ensure_future(main_routine())
    loop.run_until_complete(main_task)


@main.command(help='Create ZIP package for a template.', name='package')
@click.argument('TEMPLATE-DIR', type=DIR_TYPE, default=CURRENT_DIR, required=False)
@click.option('-o', '--output', default='template.zip', type=click.Path(writable=True),
              show_default=True, help='Target package file.')
@click.option('-f', '--force', is_flag=True, help='Delete package if already exists.')
@click.pass_context
def create_package(ctx, template_dir, output, force: bool):
    tdk = TDKCore(logger=ctx.obj.logger)
    load_local(tdk, template_dir)
    try:
        tdk.create_package(output=pathlib.Path(output), force=force)
    except Exception as e:
        ClickPrinter.failure('Failed to create the package')
        ClickPrinter.error(f'> {e}')
        sys.exit(1)
    filename = click.style(output, bold=True)
    ClickPrinter.success(f'Package {filename} created')


@main.command(help='Extract a template from ZIP package', name='unpackage')
@click.argument('TEMPLATE-PACKAGE', type=FILE_READ_TYPE, required=False)
@click.option('-o', '--output', type=NEW_DIR_TYPE, default=None, required=False,
              help='Target package file.')
@click.option('-f', '--force', is_flag=True, help='Overwrite folder if already exists.')
@click.pass_context
def extract_package(ctx, template_package, output, force: bool):
    tdk = TDKCore(logger=ctx.obj.logger)
    try:
        data = pathlib.Path(template_package).read_bytes()
        tdk.extract_package(
            zip_data=data,
            template_dir=pathlib.Path(output) if output is not None else output,
            force=force,
        )
    except Exception as e:
        ClickPrinter.failure('Failed to extract the package')
        ClickPrinter.error(f'> {e}')
        sys.exit(1)
    ClickPrinter.success(f'Package {template_package} extracted')


@main.command(help='List templates from Wizard via API.', name='list')
@click.option('-u', '--api-url', metavar='API-URL', envvar='DSW_API_URL',
              help='URL of Wizard server API.')
@click.option('-k', '--api-key', metavar='API-KEY', envvar='DSW_API_KEY',
              help='API key for Wizard instance.')
@click.option('--output-format', default=consts.DEFAULT_LIST_FORMAT,
              metavar='FORMAT', help='Entry format string for printing.')
@click.option('-r', '--released-only', is_flag=True, help='List only released templates')
@click.option('-d', '--drafts-only', is_flag=True, help='List only template drafts')
@click.pass_context
def list_templates(ctx, api_url, api_key, output_format: str,
                   released_only: bool, drafts_only: bool):
    ensure_api_config(api_url, api_key)

    async def main_routine():
        tdk = TDKCore(logger=ctx.obj.logger)
        try:
            await tdk.init_client(
                api_url=CONFIG.env.api_url,
                api_key=CONFIG.env.api_key,
            )
            if released_only:
                templates = await tdk.list_remote_templates()
                for template in templates:
                    click.echo(output_format.format(template=template))
            elif drafts_only:
                drafts = await tdk.list_remote_drafts()
                for template in drafts:
                    click.echo(output_format.format(template=template))
            else:
                click.echo('Document Templates (released)')
                templates = await tdk.list_remote_templates()
                for template in templates:
                    click.echo(output_format.format(template=template))
                click.echo('\nDocument Templates Drafts')
                drafts = await tdk.list_remote_drafts()
                for template in drafts:
                    click.echo(output_format.format(template=template))
            await tdk.safe_client.safe_close()

        except WizardCommunicationError as e:
            ClickPrinter.failure('Failed to get list of templates')
            ClickPrinter.error(f'> {e.reason}\n> {e.message}')
            await tdk.safe_client.safe_close()
            sys.exit(1)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(main_routine())


@main.command(help='Verify a template project.', name='verify')
@click.argument('TEMPLATE-DIR', type=DIR_TYPE, default=CURRENT_DIR, required=False)
@click.pass_context
def verify_template(ctx, template_dir):
    tdk = TDKCore(logger=ctx.obj.logger)
    load_local(tdk, template_dir)
    errors = tdk.verify()
    if len(errors) == 0:
        ClickPrinter.success('The template is valid!')
        print_template_info(template=tdk.safe_project.safe_template)
    else:
        ClickPrinter.failure('The template is invalid!')
        click.echo('Found violations:')
        for err in errors:
            click.echo(f' - {err.field_name}: {err.message}')


@main.group(help='Manage shared user configuration (~/.dsw-tdk).', name='config')
@click.pass_context
# pylint: disable-next=unused-argument
def config(ctx):
    pass


@config.command(name='init', help='Initialize the shared user configuration (~/.dsw-tdk).')
@click.option('-f', '--force', is_flag=True, help='Overwrite file if already exists.')
def config_init(force):
    if CONFIG.HOME_CONFIG.exists() and not force:
        ClickPrinter.error('Configuration file already exists. Use `--force` to overwrite it.')
        sys.exit(1)
    CONFIG.shared_envs.clear()
    CONFIG.default_env_name = None

    click.echo('You need to specify your first environment name (default one).')
    click.echo('Recommendation: use short and lowercase name, e.g. "production"')
    environment = click.prompt('Environment name', default='production')
    api_url = click.prompt('API URL')
    api_key = click.prompt('API Key', hide_input=True)
    try:
        CONFIG.add_shared_env(
            name=environment,
            api_url=api_url,
            api_key=api_key,
        )
        CONFIG.default_env_name = environment
    except Exception as e:
        ClickPrinter.failure('Failed to add environment')
        ClickPrinter.error(f'> {e}')
        sys.exit(1)
    while True:
        add_another = click.confirm('Do you want to add another environment?', default=False)
        if not add_another:
            break
        environment = click.prompt('Environment name')
        api_url = click.prompt('API URL')
        api_key = click.prompt('API Key', hide_input=True)
        try:
            CONFIG.add_shared_env(
                name=environment,
                api_url=api_url,
                api_key=api_key,
            )
        except Exception as e:
            ClickPrinter.failure('Failed to add environment (skipping)')
            ClickPrinter.warning(f'> {e}')

    try:
        CONFIG.persist(force=force)
        ClickPrinter.success('Configuration initialized successfully.')
    except Exception as e:
        ClickPrinter.failure('Failed to initialize configuration')
        ClickPrinter.error(f'> {e}')
        sys.exit(1)


@config.command(name='edit', help='Edit the shared user configuration (~/.dsw-tdk).')
@click.option('-f', '--force', is_flag=True, help='Create file if does not exist.')
def config_edit(force):
    if not CONFIG.HOME_CONFIG.exists():
        if force:
            CONFIG.HOME_CONFIG.parent.mkdir(parents=True, exist_ok=True)
            CONFIG.HOME_CONFIG.touch()
        else:
            ClickPrinter.error('Configuration file does not exist. Use `init` command '
                               'or `--force` flag to create it.')
            sys.exit(1)
    click.edit(
        filename=str(CONFIG.HOME_CONFIG),
        extension='.cfg',
        require_save=True,
    )


@config.command(name='check', help='Check the current configuration that can be loaded.')
def config_check():
    hidden = click.style('(hidden)', fg='red', bold=True)
    not_set = click.style('(not set)', fg='yellow', bold=True)
    click.secho('Shared configuration (~/.dsw-tdk):', bold=True)
    for env_name, env in CONFIG.shared_envs.items():
        if env_name == CONFIG.current_env_name:
            env_out = click.style(env_name, fg='green', bold=True)
            click.echo(f'{env_out} (current)')
        elif env_name == CONFIG.default_env_name:
            env_out = click.style(env_name, fg='cyan', bold=True)
            click.echo(f'{env_out} (default)')
        else:
            env_out = click.style(env_name, fg='blue', bold=True)
            click.echo(env_out)
        click.echo(f'  API URL: {env.api_url or not_set}')
        click.echo(f'  API Key: {hidden if env.api_key else not_set}')
    click.echo('')
    click.secho('Project-local configuration:', bold=True)
    if CONFIG.current_env_name == CONFIG.LOCAL_CONFIG:
        env_out = click.style(CONFIG.LOCAL_CONFIG, fg='green', bold=True)
        click.echo(f'{env_out} (current)')
    else:
        env_out = click.style('local', fg='blue', bold=True)
        click.echo(env_out)
    if CONFIG.local_env.api_url:
        click.echo(f'  API URL: {CONFIG.local_env.api_url}')
    else:
        click.echo(f'  API URL: {not_set}')
    if CONFIG.local_env.api_key:
        click.echo(f'  API Key: {hidden}')
    else:
        click.echo(f'  API Key: {not_set}')


@config.command(name='dot-env', help='Create a .env file with API configuration.')
@click.argument('TEMPLATE-DIR', type=DIR_TYPE, default=CURRENT_DIR, required=False)
@click.option('-u', '--api-url', metavar='API-URL', envvar='DSW_API_URL',
              help='URL of Wizard server API.')
@click.option('-k', '--api-key', metavar='API-KEY', envvar='DSW_API_KEY',
              help='API key for Wizard instance.')
@click.option('-f', '--force', is_flag=True, help='Overwrite file if already exists.')
@click.pass_context
def config_create_dotenv(ctx, template_dir, api_url, api_key, force):
    ensure_api_config(api_url, api_key)
    filename = ctx.obj.dot_env_file or '.env'
    output = pathlib.Path(template_dir) / filename
    try:
        if output.exists():
            if force:
                ClickPrinter.warning(f'Overwriting {output.as_posix()} (forced)')
            else:
                raise FileExistsError(f'File {output.as_posix()} already exists (not forced)')
        output.write_text(
            data=create_dot_env(
                api_url=CONFIG.env.api_url,
                api_key=CONFIG.env.api_key,
            ),
            encoding=consts.DEFAULT_ENCODING,
        )
    except Exception as e:
        ClickPrinter.failure('Failed to create dot-env file')
        ClickPrinter.error(f'> {e}')
        sys.exit(1)



# Generated file
# - do not overwrite
# - do not include in git
from collections import namedtuple

BuildInfo = namedtuple(
    'BuildInfo',
    ['version', 'built_at', 'sha', 'branch', 'tag'],
)

BUILD_INFO = BuildInfo(
    version='v4.28.4~107dd95',
    built_at='2026-03-25 16:03:18Z',
    sha='107dd9598b34665da78dd5b289fcb27488b796f0',
    branch='HEAD',
    tag='v4.28.4',
)
